import os
import pathlib
import subprocess
import tempfile
import time
from io import BytesIO

from pydub import AudioSegment

# 获取系统临时文件夹路径
temp_folder = tempfile.gettempdir()


async def wav_to_mp3(wav_data: bytes) -> bytes:
    """
    wav格式转换为mp3
    :param wav_data: wav的二进制数据
    :return: mp3的二进制数据
    """
    input_file = pathlib.Path(os.path.join(temp_folder, f"{int(time.time() * 1000)}.wav"))
    input_file.write_bytes(wav_data)
    output_file = pathlib.Path(os.path.join(temp_folder, f"{int(time.time() * 1000)}.mp3"))
    subprocess.run(["ffmpeg", "-i", input_file, "-codec:a", "libmp3lame", "-q:a", "0", output_file])
    mp3_data = output_file.read_bytes()
    try:
        os.remove(input_file)
        os.remove(output_file)
    except:
        pass

    return mp3_data


async def wav_to_mp3_2(wav_file: bytes) -> bytes:
    """
    wav格式转换为mp3
    :param wav_file: wav的二进制数据
    :return: mp3的二进制数据
    """
    audio = AudioSegment.from_file(file=BytesIO(wav_file), format='wav')
    return audio.export(format='mp3').read()



import json
import logging

from langdetect import detect
from tencentcloud.common import credential
from tencentcloud.common.exception.tencent_cloud_sdk_exception import TencentCloudSDKException
from tencentcloud.common.profile.client_profile import ClientProfile
from tencentcloud.common.profile.http_profile import HttpProfile
from tencentcloud.tmt.v20180321 import tmt_client, models

from ..config import config
from ..vits.language import Language

logging.basicConfig(level=logging.INFO)
use_tencent_tmt = False
if config.tencent_secret_key and config.tencent_secret_id:
    try:
        cred = credential.Credential(config.tencent_secret_id, config.tencent_secret_key)
        httpProfile = HttpProfile()
        httpProfile.endpoint = "tmt.tencentcloudapi.com"
        httpProfile.req
        clientProfile = ClientProfile()
        clientProfile.httpProfile = httpProfile
        client = tmt_client.TmtClient(cred, "ap-beijing", clientProfile)
        use_tencent_tmt = True
    except:
        pass


async def get_lang(text: str) -> Language | str:
    if use_tencent_tmt:
        try:
            return await _get_lang_by_tencent_tmt(text)
        except:
            return await _get_lang_by_langdetect(text)
    return await _get_lang_by_langdetect(text)


async def _get_lang_by_tencent_tmt(text: str) -> Language | str:
    try:
        # 实例化一个请求对象,每个接口都会对应一个request对象
        req = models.LanguageDetectRequest()
        params = {
            "Text": text,
            "ProjectId": 0
        }
        req.from_json_string(json.dumps(params))

        # 返回的resp是一个LanguageDetectResponse的实例，与请求对象对应
        resp = client.LanguageDetect(req)
        # 输出json格式的字符串回包
        lang = json.loads(resp.to_json_string()).get("Lang")
        if lang == 'zh':
            return Language.ZH_CN
        elif lang == 'en':
            return Language.EN
        elif lang == 'jp':
            return Language.JP
        else:
            return lang

    except TencentCloudSDKException as err:
        raise err


async def _get_lang_by_langdetect(text: str) -> Language | str:
    lang = detect(text)
    if lang == 'zh-cn':
        return Language.ZH_CN
    elif lang == 'en':
        return Language.EN
    elif lang == 'ja':
        return Language.JP
    else:
        return lang



import re

import cn2an
import opencc

converter = opencc.OpenCC('jyutjyu')

# List of (Latin alphabet, ipa) pairs:
_latin_to_ipa = [(re.compile('%s' % x[0]), x[1]) for x in [
    ('A', 'ei˥'),
    ('B', 'biː˥'),
    ('C', 'siː˥'),
    ('D', 'tiː˥'),
    ('E', 'iː˥'),
    ('F', 'e˥fuː˨˩'),
    ('G', 'tsiː˥'),
    ('H', 'ɪk̚˥tsʰyː˨˩'),
    ('I', 'ɐi˥'),
    ('J', 'tsei˥'),
    ('K', 'kʰei˥'),
    ('L', 'e˥llou˨˩'),
    ('M', 'ɛːm˥'),
    ('N', 'ɛːn˥'),
    ('O', 'ou˥'),
    ('P', 'pʰiː˥'),
    ('Q', 'kʰiːu˥'),
    ('R', 'aː˥lou˨˩'),
    ('S', 'ɛː˥siː˨˩'),
    ('T', 'tʰiː˥'),
    ('U', 'juː˥'),
    ('V', 'wiː˥'),
    ('W', 'tʊk̚˥piː˥juː˥'),
    ('X', 'ɪk̚˥siː˨˩'),
    ('Y', 'waːi˥'),
    ('Z', 'iː˨sɛːt̚˥')
]]


def number_to_cantonese(text):
    return re.sub(r'\d+(?:\.?\d+)?', lambda x: cn2an.an2cn(x.group()), text)


def latin_to_ipa(text):
    for regex, replacement in _latin_to_ipa:
        text = re.sub(regex, replacement, text)
    return text


def cantonese_to_ipa(text):
    text = number_to_cantonese(text.upper())
    text = converter.convert(text).replace('-', '').replace('$', ' ')
    text = re.sub(r'[A-Z]', lambda x: latin_to_ipa(x.group()) + ' ', text)
    text = re.sub(r'[、；：]', '，', text)
    text = re.sub(r'\s*，\s*', ', ', text)
    text = re.sub(r'\s*。\s*', '. ', text)
    text = re.sub(r'\s*？\s*', '? ', text)
    text = re.sub(r'\s*！\s*', '! ', text)
    text = re.sub(r'\s*$', '', text)
    return text



from enum import Enum


class Language(Enum):
    # 日语
    JP = "[JA]"
    # 简体中文
    ZH_CN = "[ZH]"
    # 英文
    EN = "[EN]"

    def __add__(self, other) -> str:
        return self.value + str(other)

    def __radd__(self, other) -> str:
        return str(other) + self.value



import io

import scipy.io.wavfile as wavf
import torch
from nonebot.log import logger
from torch import no_grad, LongTensor

from . import utils, commons
from .language import Language
from .models import SynthesizerTrn
from .text import text_to_sequence
from ..config import config

device = f"cuda:{config.device}" if torch.cuda.is_available() else "cpu"
logger.info(f"将使用设备{device}进行tts合成")


def get_text(text, hps, is_symbol):
    text_norm = text_to_sequence(text, hps.symbols, [] if is_symbol else hps.data.text_cleaners)
    if hps.data.add_blank:
        text_norm = commons.intersperse(text_norm, 0)
    text_norm = LongTensor(text_norm)
    return text_norm


async def generate_voice(model_path: str,
                         config_path: str,
                         language: Language,
                         text: str,
                         spk: str,
                         length_scale: float = config.default_length_scale,
                         noise_scale: float = config.default_noise_scale,
                         noise_scale_w: float = config.default_noise_scale_w) -> bytes:
    """
    生成语音，返回语音文件的bytes
    :param model_path: 模型路径
    :param config_path: 配置文件路径
    :param language: 语言
    :param text: 要合成的文本
    :param spk: 合成语音的角色名
    :param length_scale: 整体语速
    :param noise_scale: 感情变化程度
    :param noise_scale_w: 音素发音长度
    :return: wav音频文件的bytes
    """
    model_path = model_path
    config_path = config_path
    text = text
    spk = spk
    noise_scale = noise_scale
    noise_scale_w = noise_scale_w
    length_scale = length_scale

    hps = utils.get_hparams_from_file(config_path)
    net_g = SynthesizerTrn(
        len(hps.symbols),
        hps.data.filter_length // 2 + 1,
        hps.train.segment_size // hps.data.hop_length,
        n_speakers=hps.data.n_speakers,
        **hps.model).to(device)
    _ = net_g.eval()
    _ = utils.load_checkpoint(model_path, net_g, None)

    speaker_ids = hps.speakers

    text = language + text + language
    speaker_id = speaker_ids[spk]
    stn_tst = get_text(text, hps, False)
    with no_grad():
        x_tst = stn_tst.unsqueeze(0).to(device)
        x_tst_lengths = LongTensor([stn_tst.size(0)]).to(device)
        sid = LongTensor([speaker_id]).to(device)
        audio = net_g.infer(x_tst, x_tst_lengths, sid=sid, noise_scale=noise_scale, noise_scale_w=noise_scale_w,
                            length_scale=1.0 / length_scale)[0][0, 0].data.cpu().float().numpy()
    del stn_tst, x_tst, x_tst_lengths, sid

    file_obj = io.BytesIO()
    wavf.write(file_obj, hps.data.sampling_rate, audio)

    file_obj.seek(0)
    data = file_obj.read()
    with open('test.wav', mode='wb') as f:
        f.write(data)
    return data



"""
Defines the set of symbols used in text input to the model.
"""

# japanese_cleaners
# _pad        = '_'
# _punctuation = ',.!?-'
# _letters = 'AEINOQUabdefghijkmnoprstuvwyzʃʧ↓↑ '


'''# japanese_cleaners2
_pad        = '_'
_punctuation = ',.!?-~…'
_letters = 'AEINOQUabdefghijkmnoprstuvwyzʃʧʦ↓↑ '
'''

'''# korean_cleaners
_pad        = '_'
_punctuation = ',.!?…~'
_letters = 'ㄱㄴㄷㄹㅁㅂㅅㅇㅈㅊㅋㅌㅍㅎㄲㄸㅃㅆㅉㅏㅓㅗㅜㅡㅣㅐㅔ '
'''

'''# chinese_cleaners
_pad        = '_'
_punctuation = '，。！？—…'
_letters = 'ㄅㄆㄇㄈㄉㄊㄋㄌㄍㄎㄏㄐㄑㄒㄓㄔㄕㄖㄗㄘㄙㄚㄛㄜㄝㄞㄟㄠㄡㄢㄣㄤㄥㄦㄧㄨㄩˉˊˇˋ˙ '
'''

# # zh_ja_mixture_cleaners
# _pad        = '_'
# _punctuation = ',.!?-~…'
# _letters = 'AEINOQUabdefghijklmnoprstuvwyzʃʧʦɯɹəɥ⁼ʰ`→↓↑ '


'''# sanskrit_cleaners
_pad        = '_'
_punctuation = '।'
_letters = 'ँंःअआइईउऊऋएऐओऔकखगघङचछजझञटठडढणतथदधनपफबभमयरलळवशषसहऽािीुूृॄेैोौ्ॠॢ '
'''

'''# cjks_cleaners
_pad        = '_'
_punctuation = ',.!?-~…'
_letters = 'NQabdefghijklmnopstuvwxyzʃʧʥʦɯɹəɥçɸɾβŋɦː⁼ʰ`^#*=→↓↑ '
'''

'''# thai_cleaners
_pad        = '_'
_punctuation = '.!? '
_letters = 'กขฃคฆงจฉชซฌญฎฏฐฑฒณดตถทธนบปผฝพฟภมยรฤลวศษสหฬอฮฯะัาำิีึืุูเแโใไๅๆ็่้๊๋์'
'''

# # cjke_cleaners2
_pad = '_'
_punctuation = ',.!?-~…'
_letters = 'NQabdefghijklmnopstuvwxyzɑæʃʑçɯɪɔɛɹðəɫɥɸʊɾʒθβŋɦ⁼ʰ`^#*=ˈˌ→↓↑ '

'''# shanghainese_cleaners
_pad        = '_'
_punctuation = ',.!?…'
_letters = 'abdfghiklmnopstuvyzøŋȵɑɔɕəɤɦɪɿʑʔʰ̩̃ᴀᴇ15678 '
'''

'''# chinese_dialect_cleaners
_pad        = '_'
_punctuation = ',.!?~…─'
_letters = '#Nabdefghijklmnoprstuvwxyzæçøŋœȵɐɑɒɓɔɕɗɘəɚɛɜɣɤɦɪɭɯɵɷɸɻɾɿʂʅʊʋʌʏʑʔʦʮʰʷˀː˥˦˧˨˩̥̩̃̚ᴀᴇ↑↓∅ⱼ '
'''

# Export all symbols:
symbols = [_pad] + list(_punctuation) + list(_letters)

# Special symbol ids
SPACE_ID = symbols.index(" ")



import re

import opencc

dialects = {'SZ': 'suzhou', 'WX': 'wuxi', 'CZ': 'changzhou', 'HZ': 'hangzhou',
            'SX': 'shaoxing', 'NB': 'ningbo', 'JJ': 'jingjiang', 'YX': 'yixing',
            'JD': 'jiading', 'ZR': 'zhenru', 'PH': 'pinghu', 'TX': 'tongxiang',
            'JS': 'jiashan', 'HN': 'xiashi', 'LP': 'linping', 'XS': 'xiaoshan',
            'FY': 'fuyang', 'RA': 'ruao', 'CX': 'cixi', 'SM': 'sanmen',
            'TT': 'tiantai', 'WZ': 'wenzhou', 'SC': 'suichang', 'YB': 'youbu'}

converters = {}

for dialect in dialects.values():
    try:
        converters[dialect] = opencc.OpenCC(dialect)
    except:
        pass


def ngu_dialect_to_ipa(text, dialect):
    dialect = dialects[dialect]
    text = converters[dialect].convert(text).replace('-', '').replace('$', ' ')
    text = re.sub(r'[、；：]', '，', text)
    text = re.sub(r'\s*，\s*', ', ', text)
    text = re.sub(r'\s*。\s*', '. ', text)
    text = re.sub(r'\s*？\s*', '? ', text)
    text = re.sub(r'\s*！\s*', '! ', text)
    text = re.sub(r'\s*$', '', text)
    return text



""" from https://github.com/keithito/tacotron """
from nonebot.log import logger

from . import cleaners
from .symbols import symbols

# Mappings from symbol to numeric ID and vice versa:
_symbol_to_id = {s: i for i, s in enumerate(symbols)}
_id_to_symbol = {i: s for i, s in enumerate(symbols)}


def text_to_sequence(text, symbols, cleaner_names):
    """Converts a string of text to a sequence of IDs corresponding to the symbols in the text.
      Args:
        text: string to convert to a sequence
        cleaner_names: names of the cleaner functions to run the text through
      Returns:
        List of integers corresponding to the symbols in the text
    """
    sequence = []
    symbol_to_id = {s: i for i, s in enumerate(symbols)}
    clean_text = _clean_text(text, cleaner_names)
    logger.info(clean_text)
    logger.info(f" length:{len(clean_text)}")
    for symbol in clean_text:
        if symbol not in symbol_to_id.keys():
            continue
        symbol_id = symbol_to_id[symbol]
        sequence += [symbol_id]
    logger.info(f" length:{len(sequence)}")
    return sequence


def cleaned_text_to_sequence(cleaned_text, symbols):
    """Converts a string of text to a sequence of IDs corresponding to the symbols in the text.
      Args:
        text: string to convert to a sequence
      Returns:
        List of integers corresponding to the symbols in the text
    """
    symbol_to_id = {s: i for i, s in enumerate(symbols)}
    sequence = [symbol_to_id[symbol] for symbol in cleaned_text if symbol in symbol_to_id.keys()]
    return sequence


def sequence_to_text(sequence):
    """Converts a sequence of IDs back to a string"""
    result = ''
    for symbol_id in sequence:
        s = _id_to_symbol[symbol_id]
        result += s
    return result


def _clean_text(text, cleaner_names):
    for name in cleaner_names:
        cleaner = getattr(cleaners, name)
        if not cleaner:
            raise Exception('Unknown cleaner: %s' % name)
        text = cleaner(text)
    return text



from nonebot.plugin import PluginMetadata

from .config import Config
from .matcher import vits_req, help_req

__plugin_meta__ = PluginMetadata(
    name="nonebot-plugin-vits-tts",
    description="基于vits的tts语音合成",
    usage="[角色]说[要合成的文本]",
    type="application",
    homepage="https://github.com/Redmomn/nonebot-plugin-vits-tts",
    supported_adapters={"~onebot.v11"},
    config=Config,
    extra={"author": "Redmomn"}
)

__all__ = ["vits_req", "help_req"]



import json
import os
import pathlib

from nonebot.log import logger

from ..config import config


class Model:
    def __init__(self, model_path: pathlib.Path, config_path: pathlib.Path):
        self.speakers: list[str] = []
        self.model = model_path
        self.config = config_path
        with config_path.open(mode='r', encoding='utf-8') as f:
            config_data: dict = json.load(f)
        for speaker in config_data.get("speakers"):
            self.speakers.append(speaker)
        logger.info(f"已加载vits模型{model_path}")

    @property
    def model_name(self) -> str:
        return os.path.basename(os.path.dirname(self.model))

    def __str__(self) -> str:
        result = ""
        for k, v in self.__dict__.items():
            result += f"{k}:{v}\n"
        return result


def load_model(model_path: pathlib.Path) -> list[Model]:
    """
    递归加载指定文件夹的所有模型和配置文件
    模型和配置文件的文件名在config中给出
    :param model_path:
    :return: 一个 list[Model]
    """
    _models: list[Model] = []
    for root, _dir, files in os.walk(model_path):
        if (config.vmodel_file_name in files) and (config.config_file_name in files):
            _models.append(
                Model(pathlib.Path(os.path.join(root, config.vmodel_file_name)),
                      pathlib.Path(os.path.join(root, config.config_file_name))))
    return _models


def get_all_speakers_models() -> dict[str, Model]:
    """
    获取一个speaker与model实例的映射表
    :return: 一个speaker为键，model为值的dict
    """
    _speakers_models: dict[str, Model] = {}
    for _model in models:
        for _speaker in _model.speakers:
            _speakers_models[_speaker] = _model
    return _speakers_models


def get_model_from_speaker(speaker: str) -> Model | None:
    """
    从speaker的名字获取对应的model实例，不存在则返回none
    :param speaker: 说话人名称
    :return: model
    """
    return speakers_models.get(speaker, None)


models = load_model(config.vmodel_path)
speakers_models = get_all_speakers_models()
speakers = speakers_models.keys()



from nonebot import on_regex, on_command
from nonebot.adapters.onebot.v11 import Bot, GroupMessageEvent, MessageSegment
from nonebot.log import logger
from nonebot.rule import Rule
from nonebot.typing import T_State

from .rule import is_tts_msg
from .utils.audio import wav_to_mp3
from .utils.model import get_model_from_speaker, speakers
from .vits import generate_voice

vits_req = on_regex(r"^(.*)说(.*)", rule=Rule(is_tts_msg))
help_req = on_command("help", aliases={"帮助", "语音帮助", "语音合成帮助"})


@vits_req.handle()
async def _(bot: Bot, event: GroupMessageEvent, state: T_State):
    speaker = state.get("vits_speaker")
    text = state.get("vits_text")
    lang = state.get("vits_lang")

    model = get_model_from_speaker(speaker)
    logger.info(f"使用模型{model.model_name}.{speaker}生成语音：{text}")

    audio = await wav_to_mp3(await generate_voice(model_path=str(model.model),
                                                  config_path=str(model.config),
                                                  language=lang,
                                                  text=text,
                                                  spk=speaker))
    await vits_req.finish(message=MessageSegment.record(file=audio))


@help_req.handle()
async def _(bot: Bot, event: GroupMessageEvent, state: T_State):
    msg = '目前支持以下角色的语音合成：\n'
    for r in speakers:
        msg += r + ' '
    await help_req.finish(msg)



import re

from num_thai.thainumbers import NumThai

num = NumThai()

# List of (Latin alphabet, Thai) pairs:
_latin_to_thai = [(re.compile('%s' % x[0], re.IGNORECASE), x[1]) for x in [
    ('a', 'เอ'),
    ('b', 'บี'),
    ('c', 'ซี'),
    ('d', 'ดี'),
    ('e', 'อี'),
    ('f', 'เอฟ'),
    ('g', 'จี'),
    ('h', 'เอช'),
    ('i', 'ไอ'),
    ('j', 'เจ'),
    ('k', 'เค'),
    ('l', 'แอล'),
    ('m', 'เอ็ม'),
    ('n', 'เอ็น'),
    ('o', 'โอ'),
    ('p', 'พี'),
    ('q', 'คิว'),
    ('r', 'แอร์'),
    ('s', 'เอส'),
    ('t', 'ที'),
    ('u', 'ยู'),
    ('v', 'วี'),
    ('w', 'ดับเบิลยู'),
    ('x', 'เอ็กซ์'),
    ('y', 'วาย'),
    ('z', 'ซี')
]]


def num_to_thai(text):
    return re.sub(r'(?:\d+(?:,?\d+)?)+(?:\.\d+(?:,?\d+)?)?',
                  lambda x: ''.join(num.NumberToTextThai(float(x.group(0).replace(',', '')))), text)


def latin_to_thai(text):
    for regex, replacement in _latin_to_thai:
        text = re.sub(regex, replacement, text)
    return text



import re

import cn2an
import opencc

converter = opencc.OpenCC('zaonhe')

# List of (Latin alphabet, ipa) pairs:
_latin_to_ipa = [(re.compile('%s' % x[0]), x[1]) for x in [
    ('A', 'ᴇ'),
    ('B', 'bi'),
    ('C', 'si'),
    ('D', 'di'),
    ('E', 'i'),
    ('F', 'ᴇf'),
    ('G', 'dʑi'),
    ('H', 'ᴇtɕʰ'),
    ('I', 'ᴀi'),
    ('J', 'dʑᴇ'),
    ('K', 'kʰᴇ'),
    ('L', 'ᴇl'),
    ('M', 'ᴇm'),
    ('N', 'ᴇn'),
    ('O', 'o'),
    ('P', 'pʰi'),
    ('Q', 'kʰiu'),
    ('R', 'ᴀl'),
    ('S', 'ᴇs'),
    ('T', 'tʰi'),
    ('U', 'ɦiu'),
    ('V', 'vi'),
    ('W', 'dᴀbɤliu'),
    ('X', 'ᴇks'),
    ('Y', 'uᴀi'),
    ('Z', 'zᴇ')
]]


def _number_to_shanghainese(num):
    num = cn2an.an2cn(num).replace('一十', '十').replace('二十', '廿').replace('二', '两')
    return re.sub(r'((?:^|[^三四五六七八九])十|廿)两', r'\1二', num)


def number_to_shanghainese(text):
    return re.sub(r'\d+(?:\.?\d+)?', lambda x: _number_to_shanghainese(x.group()), text)


def latin_to_ipa(text):
    for regex, replacement in _latin_to_ipa:
        text = re.sub(regex, replacement, text)
    return text


def shanghainese_to_ipa(text):
    text = number_to_shanghainese(text.upper())
    text = converter.convert(text).replace('-', '').replace('$', ' ')
    text = re.sub(r'[A-Z]', lambda x: latin_to_ipa(x.group()) + ' ', text)
    text = re.sub(r'[、；：]', '，', text)
    text = re.sub(r'\s*，\s*', ', ', text)
    text = re.sub(r'\s*。\s*', '. ', text)
    text = re.sub(r'\s*？\s*', '? ', text)
    text = re.sub(r'\s*！\s*', '! ', text)
    text = re.sub(r'\s*$', '', text)
    return text



import pathlib

try:
    from nonebot import get_plugin_config
except ImportError:
    from nonebot import get_driver
from pydantic import BaseModel

try:
    from pydantic import field_validator
except ImportError:
    from pydantic import validator as field_validator


class _ScopedConfig(BaseModel):
    device: int = 0
    vmodel_path: str = 'models'
    at_bot: bool = False
    cooldown: int = 0
    vmodel_file_name: str = "model.pth"
    config_file_name: str = "config.pth"
    tencent_secret_id: str = ""
    tencent_secret_key: str = ""
    default_length_scale: float = 1
    default_noise_scale: float = .667
    default_noise_scale_w: float = .6

    @field_validator("vmodel_path")
    @classmethod
    def get_model_path(cls, value):
        return pathlib.Path(value)


class Config(BaseModel):
    vits: _ScopedConfig = _ScopedConfig()


try:
    # pydantic v2
    config = get_plugin_config(Config).vits
except:
    # pydantic v1
    config = Config.parse_obj(get_driver().config).vits



import argparse
import glob
import json
import logging
import os
import subprocess
import sys

import numpy as np
import regex as re
import torch
from scipy.io.wavfile import read

MATPLOTLIB_FLAG = False

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
logger = logging

zh_pattern = re.compile(r'[\u4e00-\u9fa5]')
en_pattern = re.compile(r'[a-zA-Z]')
jp_pattern = re.compile(r'[\u3040-\u30ff\u31f0-\u31ff]')
kr_pattern = re.compile(r'[\uac00-\ud7af\u1100-\u11ff\u3130-\u318f\ua960-\ua97f]')
num_pattern = re.compile(r'[0-9]')
comma = r"(?<=[.。!！?？；;，,、:：'\"‘“”’()（）《》「」~——])"  # 向前匹配但固定长度
tags = {'ZH': '[ZH]', 'EN': '[EN]', 'JP': '[JA]', 'KR': '[KR]'}


def tag_cjke(text):
    '''为中英日韩加tag,中日正则分不开，故先分句分离中日再识别，以应对大部分情况'''
    sentences = re.split(r"([.。!！?？；;，,、:：'\"‘“”’()（）【】《》「」~——]+ *(?![0-9]))", text)  # 分句，排除小数点
    sentences.append("")
    sentences = ["".join(i) for i in zip(sentences[0::2], sentences[1::2])]
    # print(sentences)
    prev_lang = None
    tagged_text = ""
    for s in sentences:
        # 全为符号跳过
        nu = re.sub(r'[\s\p{P}]+', '', s, flags=re.U).strip()
        if len(nu) == 0:
            continue
        s = re.sub(r'[()（）《》「」【】‘“”’]+', '', s)
        jp = re.findall(jp_pattern, s)
        # 本句含日语字符判断为日语
        if len(jp) > 0:
            prev_lang, tagged_jke = tag_jke(s, prev_lang)
            tagged_text += tagged_jke
        else:
            prev_lang, tagged_cke = tag_cke(s, prev_lang)
            tagged_text += tagged_cke
    return tagged_text


def tag_jke(text, prev_sentence=None):
    '''为英日韩加tag'''
    # 初始化标记变量
    tagged_text = ""
    prev_lang = None
    tagged = 0
    # 遍历文本
    for char in text:
        # 判断当前字符属于哪种语言
        if jp_pattern.match(char):
            lang = "JP"
        elif zh_pattern.match(char):
            lang = "JP"
        elif kr_pattern.match(char):
            lang = "KR"
        elif en_pattern.match(char):
            lang = "EN"
        # elif num_pattern.match(char):
        #     lang = prev_sentence
        else:
            lang = None
            tagged_text += char
            continue
        # 如果当前语言与上一个语言不同，就添加标记
        if lang != prev_lang:
            tagged = 1
            if prev_lang == None:  # 开头
                tagged_text = tags[lang] + tagged_text
            else:
                tagged_text = tagged_text + tags[prev_lang] + tags[lang]

            # 重置标记变量
            prev_lang = lang

        # 添加当前字符到标记文本中
        tagged_text += char

    # 在最后一个语言的结尾添加对应的标记
    if prev_lang:
        tagged_text += tags[prev_lang]
    if not tagged:
        prev_lang = prev_sentence
        tagged_text = tags[prev_lang] + tagged_text + tags[prev_lang]

    return prev_lang, tagged_text


def tag_cke(text, prev_sentence=None):
    '''为中英韩加tag'''
    # 初始化标记变量
    tagged_text = ""
    prev_lang = None
    # 是否全略过未标签
    tagged = 0

    # 遍历文本
    for char in text:
        # 判断当前字符属于哪种语言
        if zh_pattern.match(char):
            lang = "ZH"
        elif kr_pattern.match(char):
            lang = "KR"
        elif en_pattern.match(char):
            lang = "EN"
        # elif num_pattern.match(char):
        #     lang = prev_sentence
        else:
            # 略过
            lang = None
            tagged_text += char
            continue

        # 如果当前语言与上一个语言不同，添加标记
        if lang != prev_lang:
            tagged = 1
            if prev_lang == None:  # 开头
                tagged_text = tags[lang] + tagged_text
            else:
                tagged_text = tagged_text + tags[prev_lang] + tags[lang]

            # 重置标记变量
            prev_lang = lang

        # 添加当前字符到标记文本中
        tagged_text += char

    # 在最后一个语言的结尾添加对应的标记
    if prev_lang:
        tagged_text += tags[prev_lang]
    # 未标签则继承上一句标签
    if tagged == 0:
        prev_lang = prev_sentence
        tagged_text = tags[prev_lang] + tagged_text + tags[prev_lang]
    return prev_lang, tagged_text


def load_checkpoint(checkpoint_path, model, optimizer=None, drop_speaker_emb=False):
    assert os.path.isfile(checkpoint_path)
    checkpoint_dict = torch.load(checkpoint_path, map_location='cpu')
    iteration = checkpoint_dict['iteration']
    learning_rate = checkpoint_dict['learning_rate']
    if optimizer is not None:
        optimizer.load_state_dict(checkpoint_dict['optimizer'])
    saved_state_dict = checkpoint_dict['model']
    if hasattr(model, 'module'):
        state_dict = model.module.state_dict()
    else:
        state_dict = model.state_dict()
    new_state_dict = {}
    for k, v in state_dict.items():
        try:
            if k == 'emb_g.weight':
                if drop_speaker_emb:
                    new_state_dict[k] = v
                    continue
                v[:saved_state_dict[k].shape[0], :] = saved_state_dict[k]
                new_state_dict[k] = v
            else:
                new_state_dict[k] = saved_state_dict[k]
        except:
            logger.info("%s is not in the checkpoint" % k)
            new_state_dict[k] = v
    if hasattr(model, 'module'):
        model.module.load_state_dict(new_state_dict)
    else:
        model.load_state_dict(new_state_dict)
    logger.info("Loaded checkpoint '{}' (iteration {})".format(
        checkpoint_path, iteration))
    return model, optimizer, learning_rate, iteration


def save_checkpoint(model, optimizer, learning_rate, iteration, checkpoint_path):
    logger.info("Saving model and optimizer state at iteration {} to {}".format(
        iteration, checkpoint_path))
    if hasattr(model, 'module'):
        state_dict = model.module.state_dict()
    else:
        state_dict = model.state_dict()
    torch.save({'model': state_dict,
                'iteration': iteration,
                'optimizer': optimizer.state_dict() if optimizer is not None else None,
                'learning_rate': learning_rate}, checkpoint_path)


def summarize(writer, global_step, scalars={}, histograms={}, images={}, audios={}, audio_sampling_rate=22050):
    for k, v in scalars.items():
        writer.add_scalar(k, v, global_step)
    for k, v in histograms.items():
        writer.add_histogram(k, v, global_step)
    for k, v in images.items():
        writer.add_image(k, v, global_step, dataformats='HWC')
    for k, v in audios.items():
        writer.add_audio(k, v, global_step, audio_sampling_rate)


def extract_digits(f):
    digits = "".join(filter(str.isdigit, f))
    return int(digits) if digits else -1


def latest_checkpoint_path(dir_path, regex="G_[0-9]*.pth"):
    f_list = glob.glob(os.path.join(dir_path, regex))
    f_list.sort(key=lambda f: extract_digits(f))
    x = f_list[-1]
    print(f"latest_checkpoint_path:{x}")
    return x


def oldest_checkpoint_path(dir_path, regex="G_[0-9]*.pth", preserved=4):
    f_list = glob.glob(os.path.join(dir_path, regex))
    f_list.sort(key=lambda f: extract_digits(f))
    if len(f_list) > preserved:
        x = f_list[0]
        print(f"oldest_checkpoint_path:{x}")
        return x
    return ""


def plot_spectrogram_to_numpy(spectrogram):
    global MATPLOTLIB_FLAG
    if not MATPLOTLIB_FLAG:
        import matplotlib
        matplotlib.use("Agg")
        MATPLOTLIB_FLAG = True
        mpl_logger = logging.getLogger('matplotlib')
        mpl_logger.setLevel(logging.WARNING)
    import matplotlib.pylab as plt
    import numpy as np

    fig, ax = plt.subplots(figsize=(10, 2))
    im = ax.imshow(spectrogram, aspect="auto", origin="lower",
                   interpolation='none')
    plt.colorbar(im, ax=ax)
    plt.xlabel("Frames")
    plt.ylabel("Channels")
    plt.tight_layout()

    fig.canvas.draw()
    data = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep='')
    data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    plt.close()
    return data


def plot_alignment_to_numpy(alignment, info=None):
    global MATPLOTLIB_FLAG
    if not MATPLOTLIB_FLAG:
        import matplotlib
        matplotlib.use("Agg")
        MATPLOTLIB_FLAG = True
        mpl_logger = logging.getLogger('matplotlib')
        mpl_logger.setLevel(logging.WARNING)
    import matplotlib.pylab as plt
    import numpy as np

    fig, ax = plt.subplots(figsize=(6, 4))
    im = ax.imshow(alignment.transpose(), aspect='auto', origin='lower',
                   interpolation='none')
    fig.colorbar(im, ax=ax)
    xlabel = 'Decoder timestep'
    if info is not None:
        xlabel += '\n\n' + info
    plt.xlabel(xlabel)
    plt.ylabel('Encoder timestep')
    plt.tight_layout()

    fig.canvas.draw()
    data = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep='')
    data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    plt.close()
    return data


def load_wav_to_torch(full_path):
    sampling_rate, data = read(full_path)
    return torch.FloatTensor(data.astype(np.float32)), sampling_rate


def load_filepaths_and_text(filename, split="|"):
    with open(filename, encoding='utf-8') as f:
        filepaths_and_text = [line.strip().split(split) for line in f]
    return filepaths_and_text


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def get_hparams(init=True):
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--config', type=str, default="./configs/modified_finetune_speaker.json",
                        help='JSON file for configuration')
    parser.add_argument('-m', '--model', type=str, default="pretrained_models",
                        help='Model name')
    parser.add_argument('-n', '--max_epochs', type=int, default=50,
                        help='finetune epochs')
    parser.add_argument('--cont', type=str2bool, default=False,
                        help='whether to continue training on the latest checkpoint')
    parser.add_argument('--drop_speaker_embed', type=str2bool, default=False,
                        help='whether to drop existing characters')
    parser.add_argument('--train_with_pretrained_model', type=str2bool, default=True,
                        help='whether to train with pretrained model')
    parser.add_argument('--preserved', type=int, default=4,
                        help='Number of preserved models')

    args = parser.parse_args()
    model_dir = os.path.join("/", args.model)

    if not os.path.exists(model_dir):
        os.makedirs(model_dir)

    config_path = args.config
    config_save_path = os.path.join(model_dir, "config.json")
    if init:
        with open(config_path, "r") as f:
            data = f.read()
        with open(config_save_path, "w") as f:
            f.write(data)
    else:
        with open(config_save_path, "r") as f:
            data = f.read()
    config = json.loads(data)

    hparams = HParams(**config)
    hparams.model_dir = model_dir
    hparams.max_epochs = args.max_epochs
    hparams.cont = args.cont
    hparams.drop_speaker_embed = args.drop_speaker_embed
    hparams.train_with_pretrained_model = args.train_with_pretrained_model
    hparams.preserved = args.preserved
    return hparams


def get_hparams_from_dir(model_dir):
    config_save_path = os.path.join(model_dir, "config.json")
    with open(config_save_path, "r") as f:
        data = f.read()
    config = json.loads(data)

    hparams = HParams(**config)
    hparams.model_dir = model_dir
    return hparams


def get_hparams_from_file(config_path):
    with open(config_path, "r", encoding="utf-8") as f:
        data = f.read()
    config = json.loads(data)

    hparams = HParams(**config)
    return hparams


def check_git_hash(model_dir):
    source_dir = os.path.dirname(os.path.realpath(__file__))
    if not os.path.exists(os.path.join(source_dir, ".git")):
        logger.warn("{} is not a git repository, therefore hash value comparison will be ignored.".format(
            source_dir
        ))
        return

    cur_hash = subprocess.getoutput("git rev-parse HEAD")

    path = os.path.join(model_dir, "githash")
    if os.path.exists(path):
        saved_hash = open(path).read()
        if saved_hash != cur_hash:
            logger.warn("git hash values are different. {}(saved) != {}(current)".format(
                saved_hash[:8], cur_hash[:8]))
    else:
        open(path, "w").write(cur_hash)


def get_logger(model_dir, filename="train.log"):
    global logger
    logger = logging.getLogger(os.path.basename(model_dir))
    logger.setLevel(logging.DEBUG)

    formatter = logging.Formatter("%(asctime)s\t%(name)s\t%(levelname)s\t%(message)s")
    if not os.path.exists(model_dir):
        os.makedirs(model_dir)
    h = logging.FileHandler(os.path.join(model_dir, filename), encoding="utf-8")
    h.setLevel(logging.DEBUG)
    h.setFormatter(formatter)
    logger.addHandler(h)
    return logger


class HParams():
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if type(v) == dict:
                v = HParams(**v)
            self[k] = v

    def keys(self):
        return self.__dict__.keys()

    def items(self):
        return self.__dict__.items()

    def values(self):
        return self.__dict__.values()

    def __len__(self):
        return len(self.__dict__)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def __contains__(self, key):
        return key in self.__dict__

    def __repr__(self):
        return self.__dict__.__repr__()



import asyncio
import re
import time

from nonebot.adapters.onebot.v11 import Bot, GroupMessageEvent
from nonebot.typing import T_State

from .config import config
from .utils.lang_id import get_lang
from .utils.model import get_model_from_speaker
from .vits.language import Language


class GroupState:
    def __init__(self, group_id: int):
        self.lock = asyncio.Lock()
        self.group_id: int = group_id
        self.last_run_time: int = 0

    async def update_time(self):
        async with self.lock:
            self.last_run_time = time.time()

    @property
    def is_cooldown(self):
        return (int(time.time()) - self.last_run_time) < config.cooldown


class Groups:
    def __init__(self):
        self.groups: dict[int, GroupState] = {}
        self.lock = asyncio.Lock()

    async def get(self, group_id) -> GroupState:
        if not self.groups.get(group_id):
            async with self.lock:
                self.groups[group_id] = GroupState(group_id)
        return self.groups.get(group_id)


groups = Groups()


def remove_cq_at(text: str) -> str:
    pattern = re.compile(r'\[CQ:at,qq=\d+]')
    return re.sub(pattern, '', text).strip()


async def is_tts_msg(bot: Bot, event: GroupMessageEvent, state: T_State) -> bool:
    if config.at_bot and (not event.to_me):
        return False
    group_state = await groups.get(event.group_id)

    pattern = re.compile(r"^(.*)说(.*)")
    match = pattern.search(str(event.raw_message))
    if not match:
        return False
    speaker = remove_cq_at(match.group(1))
    if not get_model_from_speaker(speaker):
        await bot.send(event=event, message=f"不支持的speaker：{speaker}")
        return False
    text = match.group(2).strip()
    lang = await get_lang(text)

    if not isinstance(lang, Language):
        await bot.send(event=event, message=f"不支持的语言：{lang}")
        return False

    state['vits_speaker'] = speaker
    state['vits_text'] = text
    state['vits_lang'] = lang

    if group_state.is_cooldown:
        await bot.send(event=event, message=f"太快啦，请等待一会再发送指令")
        return False
    await group_state.update_time()
    return True



import re

from .english import english_to_lazy_ipa, english_to_ipa2
from .japanese import japanese_to_romaji_with_accent, japanese_to_ipa, japanese_to_ipa2
from .korean import latin_to_hangul, number_to_hangul, divide_hangul, korean_to_lazy_ipa, korean_to_ipa
from .mandarin import number_to_chinese, chinese_to_bopomofo, latin_to_bopomofo, chinese_to_romaji, chinese_to_lazy_ipa, \
    chinese_to_ipa
from .sanskrit import devanagari_to_ipa
from .thai import num_to_thai, latin_to_thai


def japanese_cleaners(text):
    text = japanese_to_romaji_with_accent(text)
    text = re.sub(r'([A-Za-z])$', r'\1.', text)
    return text


def japanese_cleaners2(text):
    return japanese_cleaners(text).replace('ts', 'ʦ').replace('...', '…')


def korean_cleaners(text):
    '''Pipeline for Korean text'''
    text = latin_to_hangul(text)
    text = number_to_hangul(text)
    text = divide_hangul(text)
    text = re.sub(r'([\u3131-\u3163])$', r'\1.', text)
    return text


def chinese_cleaners(text):
    '''Pipeline for Chinese text'''
    text = text.replace("[ZH]", "")
    text = number_to_chinese(text)
    text = chinese_to_bopomofo(text)
    text = latin_to_bopomofo(text)
    text = re.sub(r'([ˉˊˇˋ˙])$', r'\1。', text)
    return text


def zh_ja_mixture_cleaners(text):
    text = re.sub(r'\[ZH\](.*?)\[ZH\]',
                  lambda x: chinese_to_romaji(x.group(1)) + ' ', text)
    text = re.sub(r'\[JA\](.*?)\[JA\]', lambda x: japanese_to_romaji_with_accent(
        x.group(1)).replace('ts', 'ʦ').replace('u', 'ɯ').replace('...', '…') + ' ', text)
    text = re.sub(r'\s+$', '', text)
    text = re.sub(r'([^\.,!\?\-…~])$', r'\1.', text)
    return text


def sanskrit_cleaners(text):
    text = text.replace('॥', '।').replace('ॐ', 'ओम्')
    text = re.sub(r'([^।])$', r'\1।', text)
    return text


def cjks_cleaners(text):
    text = re.sub(r'\[ZH\](.*?)\[ZH\]',
                  lambda x: chinese_to_lazy_ipa(x.group(1)) + ' ', text)
    text = re.sub(r'\[JA\](.*?)\[JA\]',
                  lambda x: japanese_to_ipa(x.group(1)) + ' ', text)
    text = re.sub(r'\[KO\](.*?)\[KO\]',
                  lambda x: korean_to_lazy_ipa(x.group(1)) + ' ', text)
    text = re.sub(r'\[SA\](.*?)\[SA\]',
                  lambda x: devanagari_to_ipa(x.group(1)) + ' ', text)
    text = re.sub(r'\[EN\](.*?)\[EN\]',
                  lambda x: english_to_lazy_ipa(x.group(1)) + ' ', text)
    text = re.sub(r'\s+$', '', text)
    text = re.sub(r'([^\.,!\?\-…~])$', r'\1.', text)
    return text


def cjke_cleaners(text):
    text = re.sub(r'\[ZH\](.*?)\[ZH\]', lambda x: chinese_to_lazy_ipa(x.group(1)).replace(
        'ʧ', 'tʃ').replace('ʦ', 'ts').replace('ɥan', 'ɥæn') + ' ', text)
    text = re.sub(r'\[JA\](.*?)\[JA\]', lambda x: japanese_to_ipa(x.group(1)).replace('ʧ', 'tʃ').replace(
        'ʦ', 'ts').replace('ɥan', 'ɥæn').replace('ʥ', 'dz') + ' ', text)
    text = re.sub(r'\[KO\](.*?)\[KO\]',
                  lambda x: korean_to_ipa(x.group(1)) + ' ', text)
    text = re.sub(r'\[EN\](.*?)\[EN\]', lambda x: english_to_ipa2(x.group(1)).replace('ɑ', 'a').replace(
        'ɔ', 'o').replace('ɛ', 'e').replace('ɪ', 'i').replace('ʊ', 'u') + ' ', text)
    text = re.sub(r'\s+$', '', text)
    text = re.sub(r'([^\.,!\?\-…~])$', r'\1.', text)
    return text


def cjke_cleaners2(text):
    text = re.sub(r'\[ZH\](.*?)\[ZH\]',
                  lambda x: chinese_to_ipa(x.group(1)) + ' ', text)
    text = re.sub(r'\[JA\](.*?)\[JA\]',
                  lambda x: japanese_to_ipa2(x.group(1)) + ' ', text)
    text = re.sub(r'\[KO\](.*?)\[KO\]',
                  lambda x: korean_to_ipa(x.group(1)) + ' ', text)
    text = re.sub(r'\[EN\](.*?)\[EN\]',
                  lambda x: english_to_ipa2(x.group(1)) + ' ', text)
    text = re.sub(r'\s+$', '', text)
    text = re.sub(r'([^\.,!\?\-…~])$', r'\1.', text)
    return text


def thai_cleaners(text):
    text = num_to_thai(text)
    text = latin_to_thai(text)
    return text

# def shanghainese_cleaners(text):
#     text = shanghainese_to_ipa(text)
#     text = re.sub(r'([^\.,!\?\-…~])$', r'\1.', text)
#     return text


# def chinese_dialect_cleaners(text):
#     text = re.sub(r'\[ZH\](.*?)\[ZH\]',
#                   lambda x: chinese_to_ipa2(x.group(1))+' ', text)
#     text = re.sub(r'\[JA\](.*?)\[JA\]',
#                   lambda x: japanese_to_ipa3(x.group(1)).replace('Q', 'ʔ')+' ', text)
#     text = re.sub(r'\[SH\](.*?)\[SH\]', lambda x: shanghainese_to_ipa(x.group(1)).replace('1', '˥˧').replace('5',
#                   '˧˧˦').replace('6', '˩˩˧').replace('7', '˥').replace('8', '˩˨').replace('ᴀ', 'ɐ').replace('ᴇ', 'e')+' ', text)
#     text = re.sub(r'\[GD\](.*?)\[GD\]',
#                   lambda x: cantonese_to_ipa(x.group(1))+' ', text)
#     text = re.sub(r'\[EN\](.*?)\[EN\]',
#                   lambda x: english_to_lazy_ipa2(x.group(1))+' ', text)
#     text = re.sub(r'\[([A-Z]{2})\](.*?)\[\1\]', lambda x: ngu_dialect_to_ipa(x.group(2), x.group(
#         1)).replace('ʣ', 'dz').replace('ʥ', 'dʑ').replace('ʦ', 'ts').replace('ʨ', 'tɕ')+' ', text)
#     text = re.sub(r'\s+$', '', text)
#     text = re.sub(r'([^\.,!\?\-…~])$', r'\1.', text)
#     return text



from setuptools import setup, find_packages

setup(
    name='nonebot_plugin_vits_tts',
    version='0.1.1',
    packages=find_packages(),
    install_requires=[
        'Cython',
        'torch',
        'torchvision',
        'torchaudio',
        'numpy>=1.22.0',
        'scipy',
        'regex',
        'cn2an',
        'opencc',
        'inflect',
        'unidecode',
        'eng_to_ipa',
        'pyopenjtalk-prebuilt',
        'jamo',
        'ko_pron',
        'pypinyin',
        'jieba',
        'indic_transliteration',
        'num_thai',
        'pydub',
        'numba',
        'tencentcloud-sdk-python-tmt',
        'langdetect'
    ],
    author='Redmomn',
    author_email='109732988+Redmomn@users.noreply.github.com',
    description='nonebot-plugin-vits-tts',
    long_description=open('README.md', encoding='utf*8').read(),
    long_description_content_type='text/markdown',
    license='Apache-2.0',
    keywords=['nonebot2', 'vits', 'tts'],
    url='https://github.com/Redmomn/nonebot-plugin-vits-tts'
)
